import collections
import inspect
import logging
import functools
import platform
import os
import warnings
import datetime


TMP_DIR = os.path.join("/tmp", "insights-web")
logger = logging.getLogger(__name__)

TRUTH = {
    "true": True,
    "false": False,
    "yes": True,
    "no": False,
    "1": True,
    "0": False
}


def parse_bool(s, default=False):
    """
    Return the boolean value of an English string or default if it can't be
    determined.
    """
    if s is None:
        return default
    return TRUTH.get(s.lower(), default)


# python2 doesn't have a utc tzinfo by default
# see https://docs.python.org/2/library/datetime.html#tzinfo-objects
try:
    utc = datetime.timezone.utc
except:
    class UTC(datetime.tzinfo):
        """
        A tzinfo class for UTC.
        """
        ZERO = datetime.timedelta(0)

        def utcoffset(self, dt):
            return self.ZERO

        def tzname(self, dt):
            return "UTC"

        def dst(self, dt):
            return self.ZERO

    utc = UTC()


def which(cmd, env=None):
    env = env or os.environ
    if cmd.startswith("/"):
        if os.access(cmd, os.X_OK) and os.path.isfile(cmd):
            return cmd
        return None

    envpath = env.get("PATH")
    if envpath:
        paths = envpath.split(os.pathsep)
        for path in paths:
            c = os.path.join(path, cmd)
            if os.access(c, os.X_OK) and os.path.isfile(c):
                return c
    return None


class KeyPassingDefaultDict(collections.defaultdict):
    """ A default dict that passes the key to its factory function. """

    def __init__(self, *args, **kwargs):
        super(KeyPassingDefaultDict, self).__init__(*args, **kwargs)

    def __missing__(self, key):
        if self.default_factory:
            self[key] = self.default_factory(key)
            return self[key]
        else:
            return super(KeyPassingDefaultDict, self).__missing__(key)


def enum(*e):
    enums = dict(zip(e, range(len(e))))
    return type("Enum", (), enums)


def defaults(default=None):
    """
    Catches any exception thrown by the wrapped function and returns `default`
    instead.

    Parameters
    ----------

    default : object
        The default value to return if the wrapped function throws an exception
    """
    def _f(func):
        @functools.wraps(func)
        def __f(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            except Exception:
                return default
        return __f
    return _f


def keys_in(items, *args):
    """
    Use this utility function to ensure multiple keys are in one or more
    dicts.  Returns `True` if all keys are present in at least one of the
    given dicts, otherwise returns `False`.

    :Parameters:
        - `items`: Iterable of required keys
        - Variable number of subsequent arguments, each one being a dict to check.
    """

    found = dict((key, False) for key in items)
    for d in args:
        for item in items:
            if not found[item] and item in d:
                found[item] = True
    return all(found.values())


def logging_level(logger, level):
    def _f(func):
        @functools.wraps(func)
        def check_log_level(*args, **kwargs):
            if logger.getEffectiveLevel() <= level:
                return func(*args, **kwargs)

        return check_log_level

    return _f


def deprecated(func, solution, version=None):
    """
    Mark a parser or combiner as deprecated, and give a message of how to fix
    this.  This will emit a warning in the logs when the function is used.
    When combined with modifications to conftest, this causes deprecations to
    become fatal errors when testing, so they get fixed.

    Arguments:
        func (function): the function or method being deprecated.
        solution (str): a string describing the replacement class, method or
            function that replaces the thing being deprecated.  For example,
            "use the `fnord()` function" or "use the `search()` method with
            the parameter `name='(value)'`".
        version (str): The last version of insights-core that the function
            will be available before it is removed.
    """
    def get_name_line(src):
        for line in src:
            if "@" not in line:
                return line.strip()

    path = inspect.getsourcefile(func)
    src, line_no = inspect.getsourcelines(func)
    name = get_name_line(src) or "Unknown"
    if version:
        the_msg = "<{c}> at {p}:{l} is deprecated: {s} This function will be removed from version: {v}.".format(
            c=name, p=path, l=line_no, s=solution, v=version
        )
    else:
        the_msg = "<{c}> at {p}:{l} is deprecated: {s}".format(
            c=name, p=path, l=line_no, s=solution
        )

    warnings.warn(the_msg, DeprecationWarning)


def make_iter(item):
    if isinstance(item, list):
        return item
    else:
        return [item]


def ensure_dir(path, dirname=False):
    log = logging.getLogger(__name__)
    try:
        if dirname:
            path = os.path.dirname(path)
        log.debug("Ensure dir '%s'", path)
        os.makedirs(path)
    except Exception as e:
        if log.level <= logging.DEBUG:
            log.debug("Failed to ensure dir: %s", e)
        return False
    return True


def _create_log_record(msg, date, level, machine_id):
    log_record = logging.LogRecord("upload_client", logging.getLevelName(level),
                                   machine_id, None, msg.strip(), None, None)
    log_record.asctime = date
    return log_record


class objectview(object):
    def __init__(self, dict_):
        self.__dict__ = dict_


def parse_keypair_lines(content, delim='|', kv_sep='='):
    """
    Parses a set of entities, where each entity is a set of key-value pairs
    contained all on one line.  Each entity is parsed into a dictionary and
    added to the list returned from this function.
    """
    r = []
    if content:
        for row in [line for line in content if line]:
            item_dict = {}
            for item in row.split(delim):
                key, value = [i.strip("'\"").strip() for i in item.strip().split(kv_sep)]
                item_dict[key] = value
            r.append(item_dict)
    return r


def rsplit(_str, seps):
    """
    Splits _str by the first sep in seps that is found from the right side.
    Returns a tuple without the separator.
    """
    for idx, ch in enumerate(reversed(_str)):
        if ch in seps:
            return _str[0:-idx - 1], _str[-idx:]


def check_path(path):
    found = os.path.exists(path)
    logger.debug("Checking for path [%s], found = %s.", path, found)
    return found


def get_addr():
    from insights.settings import web as config

    return "http://%s:%s" % (platform.node(), config["port"])


def get_path_for_system_id(category, system_id):
    return os.path.join(TMP_DIR, category, system_id[:2], system_id)


def word_wrap(line, wrap_len=72):
    if len(line) > wrap_len:
        for i, c in enumerate(reversed(line[:wrap_len])):
            if c == " ":
                break_point = wrap_len - i
                yield line[:break_point].strip()
                for more in word_wrap(line[break_point:], wrap_len):
                    yield more
                break
    else:
        yield line.strip()


def case_variants(*elements):
    """
    For configs which take case-insensitive options, it is necessary to extend the list with
    various common case variants (all combinations are not practical). In the future, this should
    be removed, when parser filters are made case-insensitive.

    Args:
        *elements (str): list of elements which need case-sensitive expansion, you should use
                         default case such as `Ciphers`, `MACs`, `UsePAM`, `MaxAuthTries`

    Returns:
        list: list of all expanded elements
    """
    expanded_list = []
    for element in elements:
        low = element.lower()
        up = element.upper()
        title = element.title()

        # Inner case conversion, such as `MACs` or `UsePAM` to `Macs` and `UsePam`
        converted = []
        for i, letter in enumerate(element):
            if i == 0:
                converted.append(letter)
            else:
                if element[i - 1].isupper():
                    converted.append(letter.lower())
                else:
                    converted.append(letter)
        converted = "".join(converted)

        for new_element in (element, converted, low, up, title):
            if new_element not in expanded_list:
                expanded_list.append(new_element)

    return expanded_list



"""
OsRelease - file ``/etc/os-release``
====================================

This module provides plugins access to file ``/etc/os-release``.

Typical content of file ``/etc/os-release`` is::

    NAME="Red Hat Enterprise Linux Server"
    VERSION="7.2 (Maipo)"
    ID="rhel"
    ID_LIKE="fedora"
    VERSION_ID="7.2"
    PRETTY_NAME="Employee SKU"
    ANSI_COLOR="0;31"
    CPE_NAME="cpe:/o:redhat:enterprise_linux:7.2:GA:server"
    HOME_URL="https://www.redhat.com/"
    BUG_REPORT_URL="https://bugzilla.redhat.com/"

    REDHAT_BUGZILLA_PRODUCT="Red Hat Enterprise Linux 7"
    REDHAT_BUGZILLA_PRODUCT_VERSION=7.2
    REDHAT_SUPPORT_PRODUCT="Red Hat Enterprise Linux"
    REDHAT_SUPPORT_PRODUCT_VERSION="7.2"

Note:
    The /etc/os-release is not exist in RHEL6 and prior versions.

This module parses the file content and stores it as a `dict` in the `data`
attribute.

Examples:
    >>> os_rls_content = '''
    ... Red Hat Enterprise Linux Server release 7.2 (Maipo)
    ... '''.strip()
    >>> from insights.tests import context_wrap
    >>> shared = {OsRelease: OsRelease(context_wrap(os_rls_content))}
    >>> rls = shared[OsRelease]
    >>> data = rls.data
    >>> assert data.get("VARIANT_ID") is None
    >>> assert data.get("VERSION") == "7.2 (Maipo)"
"""
from .. import Parser, parser, get_active_lines, LegacyItemAccess
from insights.specs import Specs


@parser(Specs.os_release)
class OsRelease(LegacyItemAccess, Parser):
    """Parses the content of file ``/etc/os-release``."""

    def parse_content(self, content):
        data = {}
        for line in get_active_lines(content):
            k, _, v = line.partition("=")
            if _ == "=" and k:
                data[k] = v.strip('"') if v else None
        self.data = data



from insights.parsers.ls import (LSla, LSlaFiltered, LSlan, LSlanFiltered,
                                 LSlanL, LSlanR, LSlanRL, LSlanRZ, LSlanZ)
from insights.tests import context_wrap

LS_LA = """
ls: cannot access '_non_existing_': No such file or directory
/boot:
total 187380
dr-xr-xr-x.  3 root root     4096 Mar  4 16:19 .
dr-xr-xr-x. 19 root root     4096 Jul 14 09:10 ..
-rw-r--r--.  1 root root   123891 Aug 25  2015 config-3.10.0-229.14.1.el7.x86_64
drwxr-xr-x.  6 root root      111 Jun 15 09:51 grub2

/boot/grub2:
total 36
drwxr-xr-x. 6 root root  104 Mar  4 16:16 .
dr-xr-xr-x. 3 root root 4096 Mar  4 16:19 ..
lrwxrwxrwx. 1 root root   11 Aug  4  2014 menu.lst -> ./grub.cfg
-rw-r--r--. 1 root root   64 Sep 18  2015 device.map
-rw-r--r--. 1 root root 7040 Mar 29 13:30 grub.cfg
"""

LS_LAN = """
/boot:
total 187380
dr-xr-xr-x.  3 0 0     4096 Mar  4 16:19 .
dr-xr-xr-x. 19 0 0     4096 Jul 14 09:10 ..
-rw-r--r--.  1 0 0   123891 Aug 25  2015 config-3.10.0-229.14.1.el7.x86_64
drwxr-xr-x.  6 0 0      111 Jun 15 09:51 grub2

/boot/grub2:
total 36
drwxr-xr-x. 6 0 0  104 Mar  4 16:16 .
dr-xr-xr-x. 3 0 0 4096 Mar  4 16:19 ..
lrwxrwxrwx. 1 0 0   11 Aug  4  2014 menu.lst -> ./grub.cfg
-rw-r--r--. 1 0 0   64 Sep 18  2015 device.map
-rw-r--r--. 1 0 0 7040 Mar 29 13:30 grub.cfg
"""


LS_LANL = """
ls: cannot access '_non_existing_': No such file or directory
/boot:
total 187380
dr-xr-xr-x.  3 0 0     4096 Mar  4 16:19 .
dr-xr-xr-x. 19 0 0     4096 Jul 14 09:10 ..
-rw-r--r--.  1 0 0   123891 Aug 25  2015 config-3.10.0-229.14.1.el7.x86_64
drwxr-xr-x.  6 0 0      111 Jun 15 09:51 grub2

/boot/grub2:
total 32
drwx------. 4 0 0   83 Jun 27 11:01 .
dr-xr-xr-x. 5 0 0 4096 Jun  5 09:14 ..
-rw-r--r--. 1 0 0   11 Aug  4  2014 menu.lst
-rw-r--r--. 1 0 0   64 Jan 17 16:10 device.map
-rw-------. 1 0 0 6529 Jan 17 16:10 grub.cfg
"""

LS_LANR = """
ls: cannot access '_non_existing_': No such file or directory
/dev:
total 0
drwxr-xr-x. 21 0  0     3100 Jun 27 10:58 .
dr-xr-xr-x. 17 0  0      224 Apr 17 16:45 ..
crw-r--r--.  1 0  0  10, 235 Jun 27 10:58 autofs
drwxr-xr-x.  2 0  0      160 Jun 27 10:58 block
lrwxrwxrwx.  1 0  0        3 Jun 27 10:58 cdrom -> sr0

/dev/vfio:
total 0
drwxr-xr-x.  2 0 0      60 Jun 27 10:58 .
drwxr-xr-x. 21 0 0    3100 Jun 27 10:58 ..
crw-------.  1 0 0 10, 196 Jun 27 10:58 vfio

/dev/virtio-ports:
total 0
drwxr-xr-x.  2 0 0   80 Jun 27 10:58 .
drwxr-xr-x. 21 0 0 3100 Jun 27 10:58 ..
lrwxrwxrwx.  1 0 0   11 Jun 27 10:58 com.redhat.spice.0 -> ../vport2p2
lrwxrwxrwx.  1 0 0   11 Jun 27 10:58 org.qemu.guest_agent.0 -> ../vport2p1
"""

LS_LANRL = """
/dev:
total 0
drwxr-xr-x. 21    0  0            3100 Jun 27 10:58 .
dr-xr-xr-x. 17    0  0             224 Apr 17 16:45 ..
crw-r--r--.  1    0  0         10, 235 Jun 27 10:58 autofs
drwxr-xr-x.  2    0  0             160 Jun 27 10:58 block
brw-rw----.  1    0 11         11,   0 Jun 27 10:58 cdrom

/dev/vfio:
total 0
drwxr-xr-x.  2 0 0      60 Jun 27 10:58 .
drwxr-xr-x. 21 0 0    3100 Jun 27 10:58 ..
crw-------.  1 0 0 10, 196 Jun 27 10:58 vfio

/dev/virtio-ports:
total 0
drwxr-xr-x.  2 0 0     80 Jun 27 10:58 .
drwxr-xr-x. 21 0 0   3100 Jun 27 10:58 ..
crw-------.  1 0 0 243, 2 Jun 27 10:58 com.redhat.spice.0
crw-------.  1 0 0 243, 1 Jun 27 10:58 org.qemu.guest_agent.0
"""

LS_LANRZ = """
/dev:
total 0
drwxr-xr-x. 21 0  0 system_u:object_r:device_t:s0                  3100 Jun 27 10:58 .
dr-xr-xr-x. 17 0  0 system_u:object_r:root_t:s0                     224 Apr 17 16:45 ..
crw-r--r--.  1 0  0 system_u:object_r:autofs_device_t:s0            235 Jun 27 10:58 autofs
drwxr-xr-x.  2 0  0 system_u:object_r:device_t:s0                   160 Jun 27 10:58 block
lrwxrwxrwx.  1 0  0 system_u:object_r:device_t:s0                     3 Jun 27 10:58 cdrom -> sr0

/dev/vfio:
total 0
drwxr-xr-x.  2 0 0 system_u:object_r:device_t:s0           60 Jun 27 10:58 .
drwxr-xr-x. 21 0 0 system_u:object_r:device_t:s0         3100 Jun 27 10:58 ..
crw-------.  1 0 0 system_u:object_r:vfio_device_t:s0     196 Jun 27 10:58 vfio

/dev/virtio-ports:
total 0
drwxr-xr-x.  2 0 0 system_u:object_r:device_t:s0   80 Jun 27 10:58 .
drwxr-xr-x. 21 0 0 system_u:object_r:device_t:s0 3100 Jun 27 10:58 ..
lrwxrwxrwx.  1 0 0 system_u:object_r:device_t:s0   11 Jun 27 10:58 com.redhat.spice.0 -> ../vport2p2
lrwxrwxrwx.  1 0 0 system_u:object_r:device_t:s0   11 Jun 27 10:58 org.qemu.guest_agent.0 -> ../vport2p1
"""

LS_LANZ = """
/dev:
total 0
drwxr-xr-x. 21 0  0 system_u:object_r:device_t:s0          3100 Jun 27 10:58 .
dr-xr-xr-x. 17 0  0 system_u:object_r:root_t:s0             224 Apr 17 16:45 ..
crw-r--r--.  1 0  0 system_u:object_r:autofs_device_t:s0    235 Jun 27 10:58 autofs
drwxr-xr-x.  2 0  0 system_u:object_r:device_t:s0           160 Jun 27 10:58 block
lrwxrwxrwx.  1 0  0 system_u:object_r:device_t:s0             3 Jun 27 10:58 cdrom -> sr0
"""


def test_ls_la():
    ls = LSla(context_wrap(LS_LA))
    assert '/boot' in ls
    assert '/boot/grub2' in ls
    assert ls.dirs_of('/boot') == ['.', '..', 'grub2']

    grub2_files = ls.files_of('/boot/grub2')
    assert 'menu.lst' in grub2_files
    assert 'device.map' in grub2_files
    assert 'grub.cfg' in grub2_files
    assert len(grub2_files) == 3

    boot_files = ls.files_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_files
    assert len(boot_files) == 1

    boot_listings = ls.listing_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_listings
    assert boot_listings["config-3.10.0-229.14.1.el7.x86_64"]['owner'] == 'root'

    grub_listings = ls.listing_of('/boot/grub2')
    assert "menu.lst" in grub_listings
    assert grub_listings["menu.lst"]['group'] == 'root'
    assert 'link' in grub_listings["menu.lst"]
    assert grub_listings["menu.lst"]['link'] == './grub.cfg'


def test_ls_la_filtered():
    ls = LSlaFiltered(context_wrap(LS_LA))
    assert '/boot' in ls
    assert '/boot/grub2' in ls
    assert ls.dirs_of('/boot') == ['.', '..', 'grub2']

    grub2_files = ls.files_of('/boot/grub2')
    assert 'menu.lst' in grub2_files
    assert 'device.map' in grub2_files
    assert 'grub.cfg' in grub2_files
    assert len(grub2_files) == 3

    boot_files = ls.files_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_files
    assert len(boot_files) == 1

    boot_listings = ls.listing_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_listings
    assert boot_listings["config-3.10.0-229.14.1.el7.x86_64"]['owner'] == 'root'

    grub_listings = ls.listing_of('/boot/grub2')
    assert "menu.lst" in grub_listings
    assert grub_listings["menu.lst"]['group'] == 'root'
    assert 'link' in grub_listings["menu.lst"]
    assert grub_listings["menu.lst"]['link'] == './grub.cfg'


def test_ls_lan():
    ls = LSlan(context_wrap(LS_LAN))
    assert '/boot' in ls
    assert '/boot/grub2' in ls
    assert ls.dirs_of('/boot') == ['.', '..', 'grub2']

    grub2_files = ls.files_of('/boot/grub2')
    assert 'menu.lst' in grub2_files
    assert 'device.map' in grub2_files
    assert 'grub.cfg' in grub2_files
    assert len(grub2_files) == 3

    boot_files = ls.files_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_files
    assert len(boot_files) == 1

    boot_listings = ls.listing_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_listings
    assert boot_listings["config-3.10.0-229.14.1.el7.x86_64"]['owner'] == '0'

    grub_listings = ls.listing_of('/boot/grub2')
    assert "menu.lst" in grub_listings
    assert grub_listings["menu.lst"]['group'] == '0'
    assert 'link' in grub_listings["menu.lst"]
    assert grub_listings["menu.lst"]['link'] == './grub.cfg'


def test_ls_lan_filtered():
    ls = LSlanFiltered(context_wrap(LS_LAN))
    assert '/boot' in ls
    assert '/boot/grub2' in ls
    assert ls.dirs_of('/boot') == ['.', '..', 'grub2']

    grub2_files = ls.files_of('/boot/grub2')
    assert 'menu.lst' in grub2_files
    assert 'device.map' in grub2_files
    assert 'grub.cfg' in grub2_files
    assert len(grub2_files) == 3

    boot_files = ls.files_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_files
    assert len(boot_files) == 1

    boot_listings = ls.listing_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_listings
    assert boot_listings["config-3.10.0-229.14.1.el7.x86_64"]['owner'] == '0'

    grub_listings = ls.listing_of('/boot/grub2')
    assert "menu.lst" in grub_listings
    assert grub_listings["menu.lst"]['group'] == '0'
    assert 'link' in grub_listings["menu.lst"]
    assert grub_listings["menu.lst"]['link'] == './grub.cfg'


def test_ls_lanL():
    ls = LSlanL(context_wrap(LS_LANL))
    assert '/boot' in ls
    assert '/boot/grub2' in ls
    assert ls.dirs_of('/boot') == ['.', '..', 'grub2']

    grub2_files = ls.files_of('/boot/grub2')
    assert 'menu.lst' in grub2_files
    assert 'device.map' in grub2_files
    assert 'grub.cfg' in grub2_files
    assert len(grub2_files) == 3

    boot_files = ls.files_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_files
    assert len(boot_files) == 1

    boot_listings = ls.listing_of('/boot')
    assert "config-3.10.0-229.14.1.el7.x86_64" in boot_listings
    assert boot_listings["config-3.10.0-229.14.1.el7.x86_64"]['owner'] == '0'

    grub_listings = ls.listing_of('/boot/grub2')
    assert "menu.lst" in grub_listings
    assert grub_listings["menu.lst"]['group'] == '0'
    assert 'link' not in grub_listings["menu.lst"]


def test_ls_lanR():
    ls = LSlanR(context_wrap(LS_LANR))
    assert '/dev' in ls
    assert '/dev/vfio' in ls
    assert '/dev/virtio-ports' in ls
    assert ls.dirs_of('/dev') == ['.', '..', 'block']

    dev_listings = ls.listing_of('/dev')
    assert "cdrom" in dev_listings
    assert 'link' in dev_listings["cdrom"]
    assert dev_listings["cdrom"]['link'] == 'sr0'

    assert 'link' in ls.listing_of('/dev/virtio-ports')['com.redhat.spice.0']

    assert ls.files_of('non-exist') == []
    assert ls.dirs_of('non-exist') == []
    assert ls.specials_of('non-exist') == []
    assert ls.listing_of('non-exist') == []
    assert ls.total_of('non-exist') == 0
    assert ls.dir_contains('non-exist', 'test') is False
    assert ls.dir_entry('non-exist', 'test') == {}


def test_ls_lanRL():
    ls = LSlanRL(context_wrap(LS_LANRL))
    assert '/dev' in ls
    assert '/dev/vfio' in ls
    assert '/dev/virtio-ports' in ls
    assert ls.dirs_of('/dev') == ['.', '..', 'block']

    dev_listings = ls.listing_of('/dev')
    assert "cdrom" in dev_listings
    assert 'link' not in dev_listings["cdrom"]

    assert 'link' not in ls.listing_of('/dev/virtio-ports')['com.redhat.spice.0']


def test_ls_lanRZ():
    ls = LSlanRZ(context_wrap(LS_LANRZ))
    assert '/dev' in ls
    assert '/dev/vfio' in ls
    assert '/dev/virtio-ports' in ls
    assert ls.dirs_of('/dev') == ['.', '..', 'block']

    dev_listings = ls.listing_of('/dev')
    assert "cdrom" in dev_listings
    assert 'link' in dev_listings["cdrom"]
    assert dev_listings["cdrom"]['link'] == 'sr0'
    assert 'se_type' in dev_listings["cdrom"]
    assert dev_listings["cdrom"]['se_type'] == 'device_t'

    assert 'link' in ls.listing_of('/dev/virtio-ports')['com.redhat.spice.0']


def test_ls_lanZ():
    ls = LSlanZ(context_wrap(LS_LANZ))
    assert '/dev' in ls
    assert '/dev/vfio' not in ls
    assert '/dev/virtio-ports' not in ls
    assert ls.dirs_of('/dev') == ['.', '..', 'block']

    dev_listings = ls.listing_of('/dev')
    assert "cdrom" in dev_listings
    assert 'link' in dev_listings["cdrom"]
    assert dev_listings["cdrom"]['link'] == 'sr0'
    assert 'se_type' in dev_listings["cdrom"]
    assert dev_listings["cdrom"]['se_type'] == 'device_t'



"""
OdCpuDmaLatency - command ``/usr/bin/od -An -t d /dev/cpu_dma_latency``
=======================================================================

This module provides the class ``OdCpuDmaLatency`` which processes
``/usr/bin/od -An -t d /dev/cpu_dma_latency`` command output.
"""
from insights.core import CommandParser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.od_cpu_dma_latency)
class OdCpuDmaLatency(CommandParser):
    """
    Class for parsing the output of `/usr/bin/od -An -t d /dev/cpu_dma_latency` command.

    Typical output of is::

      2000000000

    Attributes:
        force_latency(int): A integer containing the value of force_latency.

    Examples:
        >>> type(cpu_dma_latency)
        <class 'insights.parsers.od_cpu_dma_latency.OdCpuDmaLatency'>
        >>> cpu_dma_latency.force_latency
        2000000000
    """

    def parse_content(self, content):
        if content and content[0].isdigit():
            self.force_latency = int(content[0])
        else:
            raise SkipComponent('Nothing to parse.')



"""
AbrtCCppConf - file "/etc/abrt/plugins/CCpp.conf"
=================================================

The AbrtCCppConf class parses the file ``/etc/abrt/plugins/CCpp.conf``.
On success it returns ``dict`` containing abrtcpp setting else it returns ``None``.

Sample Data::


    # Configuration file for CCpp hook

    # CCpp hook writes its template to the "/proc/sys/kernel/core_pattern" file
    # and stores the original template in the "/var/run/abrt/saved_core_pattern"
    # file. If you want CCpp hook to create a core dump file named according to
    # the original template as well, set 'MakeCompatCore' to 'yes'.
    # If the original template string starts with "|", the string "core" is used
    # instead of the template.
    # For more information about naming core dump files see 'man 5 core'.
    MakeCompatCore = yes

    # The option allows you to set limit for the core file size in MiB.
    #
    # This value is compared to value of the MaxCrashReportSize configuration
    # option from (/etc/abrt.conf) and the lower value is used as the limit.
    #
    # If MaxCoreFileSize is 0 then the value of MaxCrashReportSize is the limit.
    # If MaxCrashReportSize is 0 then the value of MaxCoreFileSize is the limit.
    # If both values are 0 then the core file size is unlimited.
    MaxCoreFileSize = 0

    # Do you want a copy of crashed binary be saved?
    # (useful, for example, when _deleted binary_ segfaults)
    SaveBinaryImage = no

    # When this option is set to 'yes', core backtrace is generated
    # from the memory image of the crashing process. Only the crash
    # thread is present in the backtrace.
    CreateCoreBacktrace = yes

    # Save full coredump? If set to 'no', coredump won't be saved
    # and you won't be able to report the crash to Bugzilla. Only
    # useful with CreateCoreBacktrace set to 'yes'. Please
    # note that if this option is set to 'no' and MakeCompatCore
    # is set to 'yes', the core is still written to the current
    # directory.
    SaveFullCore = yes

    # Used for debugging the hook
    #VerboseLog = 2

    # Specify where you want to store debuginfos (default: /var/cache/abrt-di)
    #
    DebuginfoLocation = /var/cache/abrt-di

    # ABRT will ignore crashes in executables whose absolute path matches one of
    # specified patterns.
    #
    #IgnoredPaths =

    # ABRT will process only crashes of either allowed users or users who are
    # members of allowed group. If no allowed users nor allowed group are specified
    # ABRT will process crashes of all users.
    #
    #AllowedUsers =
    #AllowedGroups =


Example:
    >>> type(abrt_conf)
    <class 'insights.parsers.abrt_ccpp.AbrtCCppConf'>
    >>> abrt_conf.get('CreateCoreBacktrace')
    'yes'

"""
from insights.core import Parser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.parsers import split_kv_pairs
from insights.specs import Specs


@parser(Specs.abrt_ccpp_conf)
class AbrtCCppConf(Parser, dict):
    """
    Class for parsing ``/etc/abrt/plugins/CCpp.conf``.
    """
    def parse_content(self, content):
        self.update(split_kv_pairs(content, use_partition=False))
        if not self:
            raise SkipComponent("empty content")



# -*- coding: utf-8 -*-
# Per PEP 263
import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers.systemd import unitfiles
from insights.parsers.systemd.unitfiles import ListUnits, UnitFiles
from insights.tests import context_wrap

KDUMP_DISABLED_RHEL7 = """
UNIT FILE                                   STATE
kdump.service                               disabled
""".strip()

KDUMP_ENABLED_RHEL7 = """
UNIT FILE                                   STATE
kdump.service                               enabled
""".strip()

KDUMP_DISABLED_RHEL9 = """
UNIT FILE                                   STATE           VENDOR PRESET
kdump.service                               disabled        disabled
""".strip()

KDUMP_ENABLED_RHEL9 = """
UNIT FILE                                   STATE           VENDOR PRESET
kdump.service                               enabled         enabled
""".strip()

KDUMP_ENABLED_FOOTER_RHEL7 = """
UNIT FILE                                   STATE
kdump.service                               enabled
1 unit file listed.
""".strip()

KDUMP_BIG_TEST = """
UNIT FILE                                   STATE
kdump.service                               enabled
other.service                               disabled
test.service                                static

3 unit files listed.
""".strip()

UNIT_INVALID_VS_VALID = """
UNIT FILE                                   STATE
svca.service                                enabled
svcb.service                                masked
svcc.service                                somenonsense

3 unit files listed.
"""


def test_unitfiles():
    context = context_wrap(KDUMP_DISABLED_RHEL7)
    unitfiles = UnitFiles(context)
    assert not unitfiles.is_on('kdump.service')
    assert len(unitfiles.services) == 1
    assert len(unitfiles.parsed_lines) == 1

    context = context_wrap(KDUMP_ENABLED_RHEL7)
    unitfiles = UnitFiles(context)
    assert unitfiles.is_on('kdump.service')
    assert len(unitfiles.services) == 1
    assert len(unitfiles.parsed_lines) == 1

    context = context_wrap(KDUMP_DISABLED_RHEL9)
    unitfiles = UnitFiles(context)
    assert not unitfiles.is_on('kdump.service')
    assert len(unitfiles.services) == 1
    assert len(unitfiles.parsed_lines) == 1

    context = context_wrap(KDUMP_ENABLED_RHEL9)
    unitfiles = UnitFiles(context)
    assert unitfiles.is_on('kdump.service')
    assert len(unitfiles.services) == 1
    assert len(unitfiles.parsed_lines) == 1

    context = context_wrap(KDUMP_BIG_TEST)
    unitfiles = UnitFiles(context)
    assert unitfiles.is_on('kdump.service')
    assert not unitfiles.is_on('other.service')
    assert unitfiles.is_on('test.service')
    assert len(unitfiles.services) == 3
    assert len(unitfiles.parsed_lines) == 3

    context = context_wrap(UNIT_INVALID_VS_VALID)
    unitfiles = UnitFiles(context)
    assert unitfiles.is_on('svca.service')
    assert 'svca.service' in unitfiles.services
    assert 'svca.service' in unitfiles.service_list
    assert 'svcb.service' in unitfiles.services
    assert 'svcb.service' in unitfiles.service_list
    assert 'svcc.service' not in unitfiles.services
    assert 'svcc.service' not in unitfiles.service_list
    assert True is unitfiles.is_on('svca.service')
    assert False is unitfiles.is_on('svcb.service')
    assert None is unitfiles.is_on('svcc.service')
    assert unitfiles.exists('svca.service')
    assert unitfiles.exists('svcb.service')
    assert not unitfiles.exists('svcc.service')


LISTUNITS_CONTENT = u"""
  proc-sys-fs-binfmt_misc.automount                                           loaded active waiting   Arbitrary Executable File Formats File System Automount Point
  sys-devices-pci0000:00-0000:00:03.0-virtio0-net-eth0.device                 loaded active plugged   Virtio network device
  sys-devices-pci0000:00-0000:00:04.0-virtio1-net-eth1.device                 loaded active plugged   Virtio network device
  sys-devices-pci0000:00-0000:00:05.0-virtio2-net-eth2.device                 loaded active plugged   Virtio network device
  sys-devices-pci0000:00-0000:00:07.0-virtio3-virtio\x2dports-vport3p1.device loaded active plugged   /sys/devices/pci0000:00/0000:00:07.0/virtio3/virtio-ports/vport3p1
  sys-devices-pci0000:00-0000:00:08.0-virtio4-block-vda-vda1.device           loaded active plugged   /sys/devices/pci0000:00/0000:00:08.0/virtio4/block/vda/vda1
  sys-devices-pci0000:00-0000:00:08.0-virtio4-block-vda.device                loaded active plugged   /sys/devices/pci0000:00/0000:00:08.0/virtio4/block/vda
  sys-devices-platform-serial8250-tty-ttyS1.device                            loaded active plugged   /sys/devices/platform/serial8250/tty/ttyS1
  sys-devices-platform-serial8250-tty-ttyS2.device                            loaded active plugged   /sys/devices/platform/serial8250/tty/ttyS2
  sys-devices-platform-serial8250-tty-ttyS3.device                            loaded active plugged   /sys/devices/platform/serial8250/tty/ttyS3
  sys-devices-pnp0-00:04-tty-ttyS0.device                                     loaded active plugged   /sys/devices/pnp0/00:04/tty/ttyS0
  sys-devices-virtual-block-dm\x2d0.device                                    loaded active plugged   /sys/devices/virtual/block/dm-0
  sys-devices-virtual-block-loop0.device                                      loaded active plugged   /sys/devices/virtual/block/loop0
  sys-devices-virtual-block-loop1.device                                      loaded active plugged   /sys/devices/virtual/block/loop1
  sys-devices-virtual-net-br\x2dctlplane.device                               loaded active plugged   /sys/devices/virtual/net/br-ctlplane
  sys-devices-virtual-net-br\x2dint.device                                    loaded active plugged   /sys/devices/virtual/net/br-int
  sys-devices-virtual-net-docker0.device                                      loaded active plugged   /sys/devices/virtual/net/docker0
  sys-devices-virtual-net-eth1.905.device                                     loaded active plugged   /sys/devices/virtual/net/eth1.905
  sys-devices-virtual-net-ovs\x2dsystem.device                                loaded active plugged   /sys/devices/virtual/net/ovs-system
  sys-module-configfs.device                                                  loaded active plugged   /sys/module/configfs
  sys-subsystem-net-devices-br\x2dctlplane.device                             loaded active plugged   /sys/subsystem/net/devices/br-ctlplane
  sys-subsystem-net-devices-br\x2dint.device                                  loaded active plugged   /sys/subsystem/net/devices/br-int
  sys-subsystem-net-devices-docker0.device                                    loaded active plugged   /sys/subsystem/net/devices/docker0
  sys-subsystem-net-devices-eth0.device                                       loaded active plugged   Virtio network device
  sys-subsystem-net-devices-eth1.905.device                                   loaded active plugged   /sys/subsystem/net/devices/eth1.905
  sys-subsystem-net-devices-eth1.device                                       loaded active plugged   Virtio network device
  sys-subsystem-net-devices-eth2.device                                       loaded active plugged   Virtio network device
  sys-subsystem-net-devices-ovs\x2dsystem.device                              loaded active plugged   /sys/subsystem/net/devices/ovs-system
  -.mount                                                                     loaded active mounted   /
  dev-hugepages.mount                                                         loaded active mounted   Huge Pages File System
  dev-mqueue.mount                                                            loaded active mounted   POSIX Message Queue File System
  proc-fs-nfsd.mount                                                          loaded active mounted   NFSD configuration filesystem
  run-netns.mount                                                             loaded active mounted   /run/netns
  run-user-0.mount                                                            loaded active mounted   /run/user/0
  sys-kernel-config.mount                                                     loaded active mounted   Configuration File System
  sys-kernel-debug.mount                                                      loaded active mounted   Debug File System
  var-lib-nfs-rpc_pipefs.mount                                                loaded active mounted   RPC Pipe File System
  brandbot.path                                                               loaded active waiting   Flexible branding
  systemd-ask-password-console.path                                           loaded active waiting   Dispatch Password Requests to Console Directory Watch
  systemd-ask-password-wall.path                                              loaded active waiting   Forward Password Requests to Wall Directory Watch
  session-97.scope                                                            loaded active running   Session 97 of user root
  atd.service                                                                 loaded active running   Job spooling tools
  auditd.service                                                              loaded active running   Security Auditing Service
  bootif-fix.service                                                          loaded active running   Automated fix for incorrect iPXE BOOFIF
  chronyd.service                                                             loaded active running   NTP client/server
  crond.service                                                               loaded active running   Command Scheduler
  dbus.service                                                                loaded active running   D-Bus System Message Bus
  dnsmasq.service                                                             loaded active running   DNS caching server.
  docker-distribution.service                                                 loaded active running   v2 Registry server for Docker
● docker-storage-setup.service                                                loaded failed failed    Docker Storage Setup
  docker.service                                                              loaded active running   Docker Application Container Engine
  epmd@0.0.0.0.service                                                        loaded active running   Erlang Port Mapper Daemon
  getty@tty1.service                                                          loaded active running   Getty on tty1
  gssproxy.service                                                            loaded active running   GSSAPI Proxy Daemon
  httpd.service                                                               loaded active running   The Apache HTTP Server
  iptables.service                                                            loaded active exited    IPv4 firewall with iptables
  ipxe-timeout-fix.service                                                    loaded active running   Automated fix for incorrect iPXE timeout
  ironicapi-fix.service                                                       loaded active running   Automated fix to restart ironic-api when it crashes
  irqbalance.service                                                          loaded active running   irqbalance daemon
  iscsi-shutdown.service                                                      loaded active exited    Logout off all iSCSI sessions on shutdown
  iscsid.service                                                              loaded active running   Open-iSCSI
  kdump.service                                                               loaded active exited    Crash recovery kernel arming
  kmod-static-nodes.service                                                   loaded active exited    Create list of required static device nodes for the current kernel
  ksm.service                                                                 loaded active exited    Kernel Samepage Merging
  ksmtuned.service                                                            loaded active running   Kernel Samepage Merging (KSM) Tuning Daemon
  libvirtd.service                                                            loaded active running   Virtualization daemon
  lvm2-lvmetad.service                                                        loaded active running   LVM2 metadata daemon
  lvm2-monitor.service                                                        loaded active exited    Monitoring of LVM2 mirrors, snapshots etc. using dmeventd or progress polling
  mariadb.service                                                             loaded active running   MariaDB database server
  memcached.service                                                           loaded active running   memcached daemon
  mongod.service                                                              loaded active running   High-performance, schema-free document-oriented database
  netcf-transaction.service                                                   loaded active exited    Rollback uncommitted netcf network config change transactions
  network.service                                                             loaded active exited    LSB: Bring up/down networking
● NetworkManager-wait-online.service                                          loaded failed failed    Network Manager Wait Online
  NetworkManager.service                                                      loaded active running   Network Manager
  neutron-dhcp-agent.service                                                  loaded active running   OpenStack Neutron DHCP Agent
  neutron-openvswitch-agent.service                                           loaded active running   OpenStack Neutron Open vSwitch Agent
  neutron-ovs-cleanup.service                                                 loaded active exited    OpenStack Neutron Open vSwitch Cleanup Utility
  neutron-server.service                                                      loaded active running   OpenStack Neutron Server
  openstack-aodh-evaluator.service                                            loaded active running   OpenStack Alarm evaluator service
  openstack-aodh-listener.service                                             loaded active running   OpenStack Alarm listener service
  openstack-aodh-notifier.service                                             loaded active running   OpenStack Alarm notifier service
  openstack-ceilometer-central.service                                        loaded active running   OpenStack ceilometer central agent
  openstack-ceilometer-collector.service                                      loaded active running   OpenStack ceilometer collection service
  openstack-ceilometer-notification.service                                   loaded active running   OpenStack ceilometer notification agent
  openstack-glance-api.service                                                loaded active running   OpenStack Image Service (code-named Glance) API server
  openstack-glance-registry.service                                           loaded active running   OpenStack Image Service (code-named Glance) Registry server
  openstack-heat-api-cfn.service                                              loaded active running   Openstack Heat CFN-compatible API Service
  openstack-heat-api.service                                                  loaded active running   OpenStack Heat API Service
  openstack-heat-engine.service                                               loaded active running   Openstack Heat Engine Service
  openstack-ironic-api.service                                                loaded active running   OpenStack Ironic API service
  openstack-ironic-conductor.service                                          loaded active running   OpenStack Ironic Conductor service
  openstack-ironic-inspector-dnsmasq.service                                  loaded active running   PXE boot dnsmasq service for Ironic Inspector
  openstack-ironic-inspector.service                                          loaded active running   Hardware introspection service for OpenStack Ironic
  openstack-mistral-api.service                                               loaded active running   Mistral API Server
  openstack-mistral-engine.service                                            loaded active running   Mistral Engine Server
  openstack-mistral-executor.service                                          loaded active running   Mistral Executor Server
  openstack-nova-api.service                                                  loaded active running   OpenStack Nova API Server
  openstack-nova-cert.service                                                 loaded active running   OpenStack Nova Cert Server
  openstack-nova-compute.service                                              loaded active running   OpenStack Nova Compute Server
  openstack-nova-conductor.service                                            loaded active running   OpenStack Nova Conductor Server
  openstack-nova-scheduler.service                                            loaded active running   OpenStack Nova Scheduler Server
  openstack-swift-account-reaper.service                                      loaded active running   OpenStack Object Storage (swift) - Account Reaper
  openstack-swift-account.service                                             loaded active running   OpenStack Object Storage (swift) - Account Server
  openstack-swift-container-updater.service                                   loaded active running   OpenStack Object Storage (swift) - Container Updater
  openstack-swift-container.service                                           loaded active running   OpenStack Object Storage (swift) - Container Server
  openstack-swift-object-updater.service                                      loaded active running   OpenStack Object Storage (swift) - Object Updater
  openstack-swift-object.service                                              loaded active running   OpenStack Object Storage (swift) - Object Server
  openstack-swift-proxy.service                                               loaded active running   OpenStack Object Storage (swift) - Proxy Server
  openstack-zaqar.service                                                     loaded active running   OpenStack Message Queuing Service (code-named Zaqar) Server
  openstack-zaqar@1.service                                                   loaded active running   OpenStack Message Queuing Service (code-named Zaqar) Server Instance 1
  openvswitch.service                                                         loaded active exited    Open vSwitch
● ovirt-guest-agent.service                                                   loaded failed failed    oVirt Guest Agent
  ovs-vswitchd.service                                                        loaded active running   Open vSwitch Forwarding Unit
  ovsdb-server.service                                                        loaded active running   Open vSwitch Database Unit
  polkit.service                                                              loaded active running   Authorization Manager
  postfix.service                                                             loaded active running   Postfix Mail Transport Agent
  qemu-guest-agent.service                                                    loaded active running   QEMU Guest Agent
  rabbitmq-server.service                                                     loaded active running   RabbitMQ broker
  rc-local.service                                                            loaded active exited    /etc/rc.d/rc.local Compatibility
  rhel-dmesg.service                                                          loaded active exited    Dump dmesg to /var/log/dmesg
  rhel-import-state.service                                                   loaded active exited    Import network configuration from initramfs
  rhel-push-plugin.service                                                    loaded active running   Docker Block RHEL push plugin authZ Plugin
  rhel-readonly.service                                                       loaded active exited    Configure read-only root support
  rhnsd.service                                                               loaded active running   LSB: Starts the Spacewalk Daemon
  rsyslog.service                                                             loaded active running   System Logging Service
  serial-getty@ttyS0.service                                                  loaded active running   Serial Getty on ttyS0
  sshd.service                                                                loaded active running   OpenSSH server daemon
  systemd-journal-flush.service                                               loaded active exited    Flush Journal to Persistent Storage
  systemd-journald.service                                                    loaded active running   Journal Service
  systemd-logind.service                                                      loaded active running   Login Service
  systemd-random-seed.service                                                 loaded active exited    Load/Save Random Seed
  systemd-remount-fs.service                                                  loaded active exited    Remount Root and Kernel File Systems
  systemd-sysctl.service                                                      loaded active exited    Apply Kernel Variables
  systemd-tmpfiles-setup-dev.service                                          loaded active exited    Create Static Device Nodes in /dev
  systemd-tmpfiles-setup.service                                              loaded active exited    Create Volatile Files and Directories
  systemd-udev-trigger.service                                                loaded active exited    udev Coldplug all Devices
  systemd-udevd.service                                                       loaded active running   udev Kernel Device Manager
  systemd-update-utmp.service                                                 loaded active exited    Update UTMP about System Boot/Shutdown
  systemd-user-sessions.service                                               loaded active exited    Permit User Sessions
  systemd-vconsole-setup.service                                              loaded active exited    Setup Virtual Console
  tuned.service                                                               loaded active running   Dynamic System Tuning Daemon
  xinetd.service                                                              loaded active running   Xinetd A Powerful Replacement For Inetd
  -.slice                                                                     loaded active active    Root Slice
  system-epmd.slice                                                           loaded active active    system-epmd.slice
  system-getty.slice                                                          loaded active active    system-getty.slice
  system-openstack\x2dzaqar.slice                                             loaded active active    system-openstack\x2dzaqar.slice
  system-selinux\x2dpolicy\x2dmigrate\x2dlocal\x2dchanges.slice               loaded active active    system-selinux\x2dpolicy\x2dmigrate\x2dlocal\x2dchanges.slice
  system-serial\x2dgetty.slice                                                loaded active active    system-serial\x2dgetty.slice
  system.slice                                                                loaded active active    System Slice
  user-0.slice                                                                loaded active active    user-0.slice
  user.slice                                                                  loaded active active    User and Session Slice
  dbus.socket                                                                 loaded active running   D-Bus System Message Bus Socket
  dm-event.socket                                                             loaded active listening Device-mapper event daemon FIFOs
  epmd@0.0.0.0.socket                                                         loaded active running   Erlang Port Mapper Daemon Activation Socket
  iscsid.socket                                                               loaded active running   Open-iSCSI iscsid Socket
  iscsiuio.socket                                                             loaded active listening Open-iSCSI iscsiuio Socket
  lvm2-lvmetad.socket                                                         loaded active running   LVM2 metadata daemon socket
  lvm2-lvmpolld.socket                                                        loaded active listening LVM2 poll daemon socket
  rhel-push-plugin.socket                                                     loaded active running   Docker Block RHEL push plugin Socket for the API
  rpcbind.socket                                                              loaded active listening RPCbind Server Activation Socket
  systemd-initctl.socket                                                      loaded active listening /dev/initctl Compatibility Named Pipe
  systemd-journald.socket                                                     loaded active running   Journal Socket
  systemd-shutdownd.socket                                                    loaded active listening Delayed Shutdown Socket
  systemd-udevd-control.socket                                                loaded active running   udev Control Socket
  systemd-udevd-kernel.socket                                                 loaded active running   udev Kernel Socket
  virtlockd.socket                                                            loaded active listening Virtual machine lock manager socket
  virtlogd.socket                                                             loaded active listening Virtual machine log manager socket
  swap.file.swap                                                              loaded active active    /swap.file
  basic.target                                                                loaded active active    Basic System
  cryptsetup.target                                                           loaded active active    Encrypted Volumes
  getty.target                                                                loaded active active    Login Prompts
  local-fs-pre.target                                                         loaded active active    Local File Systems (Pre)
  local-fs.target                                                             loaded active active    Local File Systems
  multi-user.target                                                           loaded active active    Multi-User System
  network-online.target                                                       loaded active active    Network is Online
  network.target                                                              loaded active active    Network
  nfs-client.target                                                           loaded active active    NFS client services
  paths.target                                                                loaded active active    Paths
  remote-fs-pre.target                                                        loaded active active    Remote File Systems (Pre)
  remote-fs.target                                                            loaded active active    Remote File Systems
  slices.target                                                               loaded active active    Slices
  sockets.target                                                              loaded active active    Sockets
  swap.target                                                                 loaded active active    Swap
  sysinit.target                                                              loaded active active    System Initialization
  timers.target                                                               loaded active active    Timers
  docker-cleanup.timer                                                        loaded active waiting   Run docker-cleanup every hour
  systemd-tmpfiles-clean.timer                                                loaded active waiting   Daily Cleanup of Temporary Directories
  unbound-anchor.timer                                                        loaded active waiting   daily update of the root trust anchor for DNSSEC
* openstack-swift-proxy.service                                               loaded failed failed    OpenStack Object Storage (swift) - Proxy Server

LOAD   = Reflects whether the unit definition was properly loaded.
ACTIVE = The high-level unit activation state, i.e. generalization of SUB.
SUB    = The low-level unit activation state, values depend on unit type.

189 loaded units listed. Pass --all to see loaded but inactive units, too.
To show all installed unit files use 'systemctl list-unit-files'.
""".strip()

LISTUNITS_CONTENT_2 = """
sockets.target                                                              loaded active active    Sockets
swap.target                                                                 loaded active active    Swap
systemd-shutdownd.socket                                                    loaded active listening Delayed Shutdown Socket
neutron-dhcp-agent.service                                                  loaded active running   OpenStack Neutron DHCP     Agent
neutron-openvswitch-agent.service                                           loaded active running   OpenStack Neutron Open     vSwitch Agent
chronyd.service                                                             loaded failed failed    NTP client/server
""".strip()


def test_listunits():
    context = context_wrap(LISTUNITS_CONTENT)
    list_units = ListUnits(context)
    service_name = "neutron-dhcp-agent.service"
    service_details = list_units.get_service_details(service_name)
    assert list_units.is_running(service_name) is True
    assert list_units.is_failed(service_name) is False
    assert service_details['UNIT'] == "neutron-dhcp-agent.service"
    service_name = "virtlogd.socket"
    service_details = list_units.get_service_details(service_name)
    assert service_details['SUB'] == "listening"
    service_name = "NetworkManager-wait-online.service"
    service_details = list_units.get_service_details(service_name)
    assert service_details['SUB'] == "failed"
    service_name = "ovirt-guest-agent.service"
    service_details = list_units.get_service_details(service_name)
    assert service_details['ACTIVE'] == "failed"
    service_name = "docker-storage-setup.service"
    assert list_units.is_running(service_name) is False
    assert list_units.is_failed(service_name) is True
    service_name = "openstack-swift-proxy.service"
    service_details = list_units.get_service_details(service_name)
    assert service_details['SUB'] == "failed"

    context = context_wrap(LISTUNITS_CONTENT_2)
    list_units = ListUnits(context)
    service_name = "systemd-shutdownd.socket"
    assert list_units.is_active(service_name) is True
    assert list_units.is_running(service_name) is False
    service_details = list_units.get_service_details(service_name)
    assert service_details['UNIT'] == "systemd-shutdownd.socket"
    service_name = "random.service"
    service_details = list_units.get_service_details(service_name)
    assert service_details['ACTIVE'] is None
    service_name = "chronyd.service"
    assert list_units.is_active(service_name) is False
    assert list_units.is_failed(service_name) is True


LISTUNITS_CONTENT_3 = u"""
UNIT                                                                                LOAD   ACTIVE SUB       DESCRIPTION
proc-sys-fs-binfmt_misc.automount                                                   loaded active waiting   Arbitrary Executable File Formats File System Automount Point
sys-devices-pci0000:00-0000:00:01.1-ata1-host0-target0:0:0-0:0:0:0-block-sr0.device loaded active plugged   QEMU_DVD-ROM
sys-devices-pci0000:00-0000:00:03.0-virtio0-net-eth0.device                         loaded active plugged   Virtio network device
sys-devices-pci0000:00-0000:00:04.0-sound-card0.device                              loaded active plugged   82801FB/FBM/FR/FW/FRW (ICH6 Family) High Definition Audio Controller (QEMU Virtual Machine)
sys-devices-pci0000:00-0000:00:06.0-virtio1-virtio\x2dports-vport1p1.device         loaded active plugged   /sys/devices/pci0000:00/0000:00:06.0/virtio1/virtio-ports/vport1p1
sys-devices-pci0000:00-0000:00:06.0-virtio1-virtio\x2dports-vport1p2.device         loaded active plugged   /sys/devices/pci0000:00/0000:00:06.0/virtio1/virtio-ports/vport1p2
sys-devices-pci0000:00-0000:00:07.0-virtio2-block-vda-vda1.device                   loaded active plugged   /sys/devices/pci0000:00/0000:00:07.0/virtio2/block/vda/vda1
sys-devices-pci0000:00-0000:00:07.0-virtio2-block-vda-vda2.device                   loaded active plugged   LVM PV m1dNfE-0R8d-dbBp-n4F9-mT11-dffC-M4S0wq on /dev/vda2 2
sys-devices-pci0000:00-0000:00:07.0-virtio2-block-vda.device                        loaded active plugged   /sys/devices/pci0000:00/0000:00:07.0/virtio2/block/vda
insights-client.timer                                                               loaded active waiting   Insights Client Timer Task
systemd-tmpfiles-clean.timer                                                        loaded active waiting   Daily Cleanup of Temporary Directories
unbound-anchor.timer                                                                loaded active waiting   daily update of the root trust anchor for DNSSEC

LOAD   = Reflects whether the unit definition was properly loaded.
ACTIVE = The high-level unit activation state, i.e. generalization of SUB.
SUB    = The low-level unit activation state, values depend on unit type.

161 loaded units listed. Pass --all to see loaded but inactive units, too.
To show all installed unit files use 'systemctl list-unit-files'.
""".strip()


def test_listunits_headings():
    list_units = ListUnits(context_wrap(LISTUNITS_CONTENT_3))
    assert set(list_units.service_names) == set([
        'proc-sys-fs-binfmt_misc.automount',
        'sys-devices-pci0000:00-0000:00:01.1-ata1-host0-target0:0:0-0:0:0:0-block-sr0.device',
        'sys-devices-pci0000:00-0000:00:03.0-virtio0-net-eth0.device',
        'sys-devices-pci0000:00-0000:00:04.0-sound-card0.device',
        u'sys-devices-pci0000:00-0000:00:06.0-virtio1-virtio\x2dports-vport1p1.device',
        u'sys-devices-pci0000:00-0000:00:06.0-virtio1-virtio\x2dports-vport1p2.device',
        'sys-devices-pci0000:00-0000:00:07.0-virtio2-block-vda-vda1.device',
        'sys-devices-pci0000:00-0000:00:07.0-virtio2-block-vda-vda2.device',
        'sys-devices-pci0000:00-0000:00:07.0-virtio2-block-vda.device',
        'insights-client.timer',
        'systemd-tmpfiles-clean.timer',
        'unbound-anchor.timer'
    ])
    assert list_units.get_service_details('insights-client.timer')['DESCRIPTION'] == 'Insights Client Timer Task'


LISTUNITS_DOCTEST = """
UNIT                                LOAD   ACTIVE SUB       DESCRIPTION
sockets.target                      loaded active active    Sockets
swap.target                         loaded active active    Swap
systemd-shutdownd.socket            loaded active listening Delayed Shutdown Socket
neutron-dhcp-agent.service          loaded active running   OpenStack Neutron DHCP Agent
neutron-openvswitch-agent.service   loaded active running   OpenStack Neutron Open vSwitch Agent
...
unbound-anchor.timer                loaded active waiting   daily update of the root trust anchor for DNSSEC

LOAD   = Reflects whether the unit definition was properly loaded.
ACTIVE = The high-level unit activation state, i.e. generalization of SUB.
SUB    = The low-level unit activation state, values depend on unit type.

161 loaded units listed. Pass --all to see loaded but inactive units, too.
To show all installed unit files use 'systemctl list-unit-files'.
""".strip()


UNITFILES_DOCTEST = """
UNIT FILE                                     STATE
mariadb.service                               enabled
neutron-openvswitch-agent.service             enabled
neutron-ovs-cleanup.service                   enabled
neutron-server.service                        enabled
runlevel0.target                              disabled
runlevel1.target                              disabled
runlevel2.target                              enabled
""".strip()


def test_unitfiles_doc_examples():
    env = {
        'conf': UnitFiles(context_wrap(UNITFILES_DOCTEST)),
        'units': ListUnits(context_wrap(LISTUNITS_DOCTEST)),
    }
    failed, total = doctest.testmod(unitfiles, globs=env)
    assert failed == 0


UNITFILES_NG = """
Failed to list unit files: Connection timed out
""".strip()

LISTUNITS_NG = """
Failed to list units
""".strip()


def test_unitfile_NG():
    with pytest.raises(SkipComponent):
        UnitFiles(context_wrap(UNITFILES_NG))

    with pytest.raises(SkipComponent):
        ListUnits(context_wrap(LISTUNITS_NG))



"""
SystemctlShow - command ``systemctl show``
==========================================

Parsers included in this module are:

SystemctlShowServiceAll - command ``systemctl show *.service``
--------------------------------------------------------------
SystemctlShowTarget - command ``systemctl show *.target``
---------------------------------------------------------
SystemctlShowAllServiceWithLimitedProperties - command ``systemctl show *.service --all --property=<...>``
----------------------------------------------------------------------------------------------------------
"""
from insights.core import CommandParser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.parsers import split_kv_pairs
from insights.specs import Specs


@parser(Specs.systemctl_show_all_services)
class SystemctlShowServiceAll(CommandParser, dict):
    """
    Class for parsing ``systemctl show *.service`` command output.
    Empty properties are suppressed.

    Sample Input::

        Id=postfix.service
        Names=postfix.service
        TimeoutStartUSec=1min 30s
        LimitNOFILE=65536
        LimitMEMLOCK=
        LimitLOCKS=18446744073709551615

        Id=postgresql.service
        Names=postgresql.service
        Requires=basic.target
        LimitMSGQUEUE=819200
        LimitNICE=0

    Sample Output::

        {
            "postfix.service": {
                "Id"               : "postfix.service",
                "Names"            : "postfix.service",
                "LimitNOFILE"      : "65536",
                "TimeoutStartUSec" : "1min 30s",
                "LimitLOCKS"       : "18446744073709551615",
            },
            "postgresql.service": {
                "Id"               : "postgresql.service",
                "Names"            : "postgresql.service",
                "Requires"         : "basic.target",
                "LimitMSGQUEUE"    : "819200",
                "LimitNICE"        : "0",
            }
        }

    Examples:
        >>> 'postfix' in systemctl_show_all  # ".service" is needed
        False
        >>> 'postfix.service' in systemctl_show_all
        True
        >>> systemctl_show_all['postfix.service']['Id']
        'postfix.service'
        >>> 'LimitMEMLOCK' in systemctl_show_all['postfix.service']
        False
        >>> systemctl_show_all['postfix.service']['LimitLOCKS']
        '18446744073709551615'
        >>> 'postgresql.service' in systemctl_show_all
        True
        >>> systemctl_show_all['postgresql.service']['LimitNICE']
        '0'

    Raises:
        SkipComponent: When nothing needs to parse
        ParseException: When something cannot be parsed
    """

    def parse_content(self, content):
        if not content:
            raise SkipComponent

        sidx = 0
        idx_list = []
        for i, l in enumerate(content):
            if l.strip() == '':
                idx_list.append((sidx, i))
                sidx = i + 1
        idx_list.append((sidx, len(content)))
        for s, e in idx_list:
            data = split_kv_pairs(content[s:e], use_partition=False)
            name = data.get('Names', data.get('Id'))
            if not name:
                raise ParseException('"Names" or "Id" not found!')
            self[name] = dict((k, v) for k, v in data.items() if v)

        if len(self) == 0:
            raise SkipComponent


@parser(Specs.systemctl_show_target)
class SystemctlShowTarget(SystemctlShowServiceAll):
    """
    Class for parsing ``systemctl show *.target`` command output.
    Empty properties are suppressed.

    This class is inherited from :py:class:`SystemctlShowServiceAll`.

    Sample Input::

        Id=network.target
        Names=network.target
        WantedBy=NetworkManager.service
        Conflicts=shutdown.target
        Before=tuned.service network-online.target rhsmcertd.service kdump.service httpd.service rsyslog.service rc-local.service insights-client.timer insights-client.service sshd.service postfix.service
        After=firewalld.service network-pre.target network.service NetworkManager.service
        Documentation=man:systemd.special(7) http://www.freedesktop.org/wiki/Software/systemd/NetworkTarget
        Description=Network
        LoadState=loaded
        ActiveState=active
        SubState=active
        FragmentPath=/usr/lib/systemd/system/network.target
        UnitFileState=static
        UnitFilePreset=disabled
        InactiveExitTimestamp=Tue 2020-02-25 10:39:46 GMT
        InactiveExitTimestampMonotonic=15332468
        ActiveEnterTimestamp=Tue 2020-02-25 10:39:46 GMT
        ActiveEnterTimestampMonotonic=15332468
        ActiveExitTimestampMonotonic=0
        InactiveEnterTimestampMonotonic=0
        CanStart=no

    Sample Output::

        {'network.target': {'ActiveEnterTimestamp': 'Tue 2020-02-25 10:39:46 GMT',
                            'ActiveEnterTimestampMonotonic': '15332468',
                            'ActiveExitTimestampMonotonic': '0',
                            'ActiveState': 'active',
                            'After': 'firewalld.service network-pre.target '
                                     'network.service NetworkManager.service',
                            'Before': 'tuned.service network-online.target '
                                      'rhsmcertd.service kdump.service httpd.service '
                                      'rsyslog.service rc-local.service '
                                      'insights-client.timer insights-client.service '
                                      'sshd.service postfix.service',
                            'CanStart': 'no',
                            'Conflicts': 'shutdown.target',
                            'Description': 'Network',
                            'Documentation': 'man:systemd.special(7) '
                                             'http://www.freedesktop.org/wiki/Software/systemd/NetworkTarget',
                            'FragmentPath': '/usr/lib/systemd/system/network.target',
                            'Id': 'network.target',
                            'InactiveEnterTimestampMonotonic': '0',
                            'InactiveExitTimestamp': 'Tue 2020-02-25 10:39:46 GMT',
                            'InactiveExitTimestampMonotonic': '15332468',
                            'LoadState': 'loaded',
                            'Names': 'network.target',
                            'SubState': 'active',
                            'UnitFilePreset': 'disabled',
                            'UnitFileState': 'static',
                            'WantedBy': 'NetworkManager.service'})

    Examples:
        >>> 'network.target' in systemctl_show_target
        True
        >>> systemctl_show_target.get('network.target').get('WantedBy', None)
        'NetworkManager.service'
        >>> systemctl_show_target.get('network.target').get('RequiredBy', None)

    Raises:
        SkipComponent: When nothing needs to parse
        ParseException: When something cannot be parsed
    """
    pass


@parser(Specs.systemctl_show_all_services_with_limited_properties)
class SystemctlShowAllServiceWithLimitedProperties(SystemctlShowServiceAll):
    """
    Class for parsing the output of command ``systemctl show *.service --all --property=<...>``.

    Sample Input::

        CPUAccounting=no
        CPUShares=18446744073709551615
        StartupCPUShares=18446744073709551615
        CPUQuotaPerSecUSec=infinity
        Names=test1.service
        SubState=dead
        UnitFileState=static

        CPUAccounting=yes
        CPUShares=18446744073709551615
        StartupCPUShares=18446744073709551615
        CPUQuotaPerSecUSec=infinity
        Names=test2.service
        SubState=dead
        UnitFileState=enabled

    Sample Output::

        {'test1.service': {
            'Names': 'test1.service',
            'SubState': 'dead',
            'UnitFileState': 'static',
            'CPUQuotaPerSecUSec': 'infinity',
            'CPUShares': '18446744073709551615',
            'StartupCPUShares': '18446744073709551615',
            'CPUAccounting': 'no'
            },
        'test2.service': {
            'Names': 'test2.service',
            'SubState': 'dead',
            'UnitFileState': 'enabled',
            'CPUQuotaPerSecUSec': 'infinity',
            'CPUShares': '18446744073709551615',
            'StartupCPUShares': '18446744073709551615',
            'CPUAccounting': 'yes'
            }
        }

    Examples:
        >>> 'test1.service' in all_services_with_limited_info
        True
        >>> all_services_with_limited_info.get('test1.service').get('CPUAccounting', None)
        'no'
        >>> 'test2.service' in all_services_with_limited_info.get_services_with_cpuaccouting_enabled()
        True
    """

    def get_services_with_cpuaccouting_enabled(self):
        """Return a list of service names which enables the CPUAccounting."""
        services_with_cpuaccouting_enabled = []
        for service_name, service_info in self.items():
            service_enabled_or_stared = (
                service_info.get('UnitFileState') in ['static', 'enabled'] or
                service_info.get('SubState') in ['running', 'failed', 'dead', 'exited'])
            cpuaccounting_enabled = (
                service_info.get('CPUAccounting', '').lower() == 'yes' or
                service_info.get('CPUQuotaPerSecUSec', '').lower() != 'infinity')
            val1 = service_info.get('CPUShares')
            val2 = service_info.get('StartupCPUShares')
            # From the man page of systemctl, the value of CPUShares is the range 2 and 262144
            if any(v.isdigit() and 2 <= int(v) <= 262144 for v in (val1, val2)):
                cpuaccounting_enabled = True
            if service_enabled_or_stared and cpuaccounting_enabled:
                services_with_cpuaccouting_enabled.append(service_name)
        return services_with_cpuaccouting_enabled



import doctest

from insights.parsers import ls_var_lib_mongodb
from insights.parsers.ls_var_lib_mongodb import LsVarLibMongodb
from insights.tests import context_wrap

LS_VAR_LIB_MONGODB = """
total 6322200
drwxr-xr-x.  3 mongodb mongodb        256 Jun  7 10:07 .
drwxr-xr-x. 71 root    root          4096 Jun 22 10:35 ..
drwxr-xr-x.  2 mongodb mongodb         65 Jul 10 09:33 journal
-rw-------.  1 mongodb mongodb   67108864 Jul 10 09:32 local.0
-rw-------.  1 mongodb mongodb   16777216 Jul 10 09:32 local.ns
"""


def test_ls_var_lib_mongodb():
    ls_var_lib_mongodb = LsVarLibMongodb(context_wrap(LS_VAR_LIB_MONGODB, path="insights_commands/ls_-la_.var.lib.mongodb"))
    assert ls_var_lib_mongodb.dirs_of('/var/lib/mongodb') == ['.', '..', 'journal']
    journal = ls_var_lib_mongodb.dir_entry('/var/lib/mongodb', 'journal')
    assert journal is not None
    assert journal == {'group': 'mongodb', 'name': 'journal', 'links': 2, 'perms': 'rwxr-xr-x.', 'raw_entry': 'drwxr-xr-x.  2 mongodb mongodb         65 Jul 10 09:33 journal', 'owner': 'mongodb', 'date': 'Jul 10 09:33', 'type': 'd', 'dir': '/var/lib/mongodb', 'size': 65}


def test_ls_var_lib_mongodb_doc_examples():
    env = {
        'LsVarLibMongodb': LsVarLibMongodb,
        'ls_var_lib_mongodb': LsVarLibMongodb(context_wrap(LS_VAR_LIB_MONGODB, path="insights_commands/ls_-la_.var.lib.mongodb")),
    }
    failed, total = doctest.testmod(ls_var_lib_mongodb, globs=env)
    assert failed == 0



from insights.parsers import satellite_installer_configurations
from insights.tests import context_wrap
import doctest

CUSTOM_HIERA_CONFIG = """
mongodb::server::config_data:
  storage.wiredTiger.engineConfig.cacheSizeGB: 8
apache::mod::prefork::serverlimit: 582
apache::mod::prefork::maxclients: 582
apache::mod::prefork::startservers: 10
"""

CUSTOM_HIERA_CONFIG_2 = """
---
# This YAML file lets you set your own custom configuration in Hiera for the
# installer puppet modules that might not be exposed to users directly through
# installer arguments.
#
# For example, to set 'TraceEnable Off' in Apache, a common requirement for
# security auditors, add this to this file:
#
#   apache::trace_enable: Off
#
# Consult the full module documentation on http://forge.puppetlabs.com,
# or the actual puppet classes themselves, to discover options to configure.
#
# Do note, setting some values may have unintended consequences that affect the
# performance or functionality of the application. Consider the impact of your
# changes before applying them, and test them in a non-production environment
# first.
#
# Here are some examples of how you tune the Apache options if needed:
#
# apache::mod::prefork::startservers: 8

"""


def test_custom_hiera():
    result = satellite_installer_configurations.CustomHiera(context_wrap(CUSTOM_HIERA_CONFIG))
    assert result.data['mongodb::server::config_data']['storage.wiredTiger.engineConfig.cacheSizeGB'] == 8
    assert result.data['apache::mod::prefork::startservers'] == 10
    assert result.data['apache::mod::prefork::maxclients'] == 582

    result = satellite_installer_configurations.CustomHiera(context_wrap(CUSTOM_HIERA_CONFIG_2))
    assert result.data is None


def test_doc():
    env = {
        'custom_hiera': satellite_installer_configurations.CustomHiera(context_wrap(CUSTOM_HIERA_CONFIG))
    }
    failed, total = doctest.testmod(satellite_installer_configurations, globs=env)
    assert failed == 0



# -*- coding: UTF-8 -*-
from insights.core import SysconfigOptions
from insights.tests import context_wrap


STANDARD_CONFIG = """
# This is a standard, but made up, config file example
OPTIONS="-x -g"

USER=root
HOST=192.168.0.100
PORT=1066
"""


TRICKY_CONFIG = """
# COMMENTED_OUT=1
SELECTOR='and all that'
     SPACES_IN_FRONT=yes
COMMENT_SPACE_OK=value_with_#
SPACE_COMMENT_IGNORED=value_with #nothing else
COMMENT_NOSPACE_OK=value_with_#_in_it
COMMENT_QUOTED_IN_VALUE="This value should have a # character in it"
MANY_QUOTES="this "'is'" valid"
QUOTE_HANDLING="single ' allowed"
OPTION_WITH_NO_VALUE=
DOUBLE_EQUALS==included
  # blank lines with a comment are ignored

# blank lines without comments are ignored too.
# Comments with apostrophes shouldn't fool the dequoting process
VALUE_ON_QUOTE_COMMENT_LINE=this_is_OK # but this hasn't been read
this line should be in unparsed lines afterward
AND = 'so should this'
""" + "SPACES_AFTER_VALUE=should_be_ignored    \n" + "  \n"


def test_standard_config():
    config = SysconfigOptions(context_wrap(STANDARD_CONFIG))

    assert 'OPTIONS' in config.data
    assert 'USER' in config.data
    assert 'HOST' in config.data
    assert 'PORT' in config.data

    assert config.data['OPTIONS'] == '-x -g'
    assert config.data['USER'] == 'root'
    assert config.data['HOST'] == '192.168.0.100'
    assert config.data['PORT'] == '1066'

    # New pseudo-dictionary access:
    assert sorted(config.keys()) == sorted(['OPTIONS', 'USER', 'HOST', 'PORT'])
    assert 'OPTIONS' in config
    assert 'USER' in config
    assert 'HOST' in config
    assert 'PORT' in config

    assert config['OPTIONS'] == '-x -g'
    assert config['USER'] == 'root'
    assert config['HOST'] == '192.168.0.100'
    assert config['PORT'] == '1066'

    assert config.get('OPTIONS', 'wrong') == '-x -g'
    # Check that get for not found value returns default
    assert config.get('NO_KEY', 'value') == 'value'


def test_tricky_config():
    config = SysconfigOptions(context_wrap(TRICKY_CONFIG))

    assert sorted(config.data.keys()) == sorted([
        'SELECTOR', 'SPACES_IN_FRONT', 'SPACES_AFTER_VALUE',
        'COMMENT_SPACE_OK', 'SPACE_COMMENT_IGNORED', 'COMMENT_NOSPACE_OK',
        'COMMENT_QUOTED_IN_VALUE', 'MANY_QUOTES', 'QUOTE_HANDLING',
        'OPTION_WITH_NO_VALUE', 'DOUBLE_EQUALS',
        'VALUE_ON_QUOTE_COMMENT_LINE'
    ])

    assert config.data['SELECTOR'] == 'and all that'
    assert config.data['SPACES_IN_FRONT'] == 'yes'
    assert config.data['SPACES_AFTER_VALUE'] == 'should_be_ignored'
    assert config.data['COMMENT_SPACE_OK'] == 'value_with_#'
    assert config.data['SPACE_COMMENT_IGNORED'] == 'value_with'
    assert config.data['COMMENT_NOSPACE_OK'] == 'value_with_#_in_it'
    assert config.data['COMMENT_QUOTED_IN_VALUE'] == "This value should have a # character in it"
    assert config.data['MANY_QUOTES'] == "this is valid"
    assert config.data['QUOTE_HANDLING'] == "single ' allowed"
    assert config.data['OPTION_WITH_NO_VALUE'] == ''
    assert config.data['DOUBLE_EQUALS'] == '=included'
    assert config.data['VALUE_ON_QUOTE_COMMENT_LINE'] == 'this_is_OK'

    assert config.unparsed_lines == [
        'this line should be in unparsed lines afterward',
        "AND = 'so should this'",
    ]



"""
Custom datasources related to ``httpd``
"""
import json

from insights.combiners.ps import Ps
from insights.core.context import HostContext
from insights.core.exceptions import SkipComponent
from insights.core.plugins import datasource
from insights.core.spec_factory import DatasourceProvider
from insights.parsers.mount import ProcMounts
from insights.specs.datasources import get_running_commands


@datasource(Ps, HostContext)
def httpd_cmds(broker):
    """
    Function to search the output of ``ps auxcww`` to find all running Apache
    webserver processes and extract the binary path.

    Returns:
        list: List of the binary paths to each running process
    """
    cmds = get_running_commands(broker[Ps], broker[HostContext], ['httpd', ])
    if cmds:
        return cmds
    raise SkipComponent


@datasource(ProcMounts, HostContext)
def httpd_on_nfs(broker):
    """
    Function to get the count of httpd opened file on nfs v4

    Returns:
        str: JSON string with keys: "httpd_ids", "nfs_mounts", "open_nfs_files"
    """
    mnt = broker[ProcMounts]
    mps = mnt.search(mount_type='nfs4')
    # get nfs 4.0 mount points
    nfs_mounts = [m.mount_point for m in mps if 'vers' in m.mount_options and m.mount_options.vers.startswith("4")]
    if nfs_mounts:
        # get all httpd ps
        httpd_pids = broker[HostContext].shell_out("pgrep httpd")
        if httpd_pids:
            open_nfs_files = 0
            lsof_cmds = ["lsof -p {0}".format(pid) for pid in httpd_pids if pid]
            # maybe there are thousands open files
            for cmd in lsof_cmds:
                for line in broker[HostContext].shell_out(cmd):
                    items = line.split()
                    if len(items) > 8 and items[8].startswith(tuple(nfs_mounts)):
                        open_nfs_files += 1
            result_dict = {"http_ids": httpd_pids, "nfs_mounts": nfs_mounts, "open_nfs_files": open_nfs_files}
            relative_path = 'insights_commands/httpd_open_nfsV4_files'
            return DatasourceProvider(content=json.dumps(result_dict), relative_path=relative_path)
    raise SkipComponent



import doctest
from insights.parsers import ls_var_lib_nova_instances as ls_instances
from insights.tests import context_wrap


LS_VAR_LIB_NOVA_INSTANCES = '''
/var/lib/nova/instances/:
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 .
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 ..
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 11415c6c-a2a5-45f0-a198-724246b96631
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 _base
-rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 compute_nodes
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 locks

/var/lib/nova/instances/11415c6c-a2a5-45f0-a198-724246b96631:
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 .
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 ..
-rw-------. root root system_u:object_r:nova_var_lib_t:s0 console.log
-rw-r--r--. qemu qemu system_u:object_r:svirt_image_t:s0:c92,c808 disk
-rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 disk.info

/var/lib/nova/instances/_base:
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 .
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 ..
-rw-r--r--. qemu qemu system_u:object_r:virt_content_t:s0 572dfdb7e1d9304342cbe1fd5e3da4ff2e55c7a6

/var/lib/nova/instances/locks:
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 .
drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 ..
-rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 nova-572dfdb7e1d9304342cbe1fd5e3da4ff2e55c7a6
-rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 nova-storage-registry-lock
'''.strip()


LS_R_VAR_LIB_NOVA_INSTANCES = '''
/var/lib/nova/instances:
total 4
drwxr-xr-x. 5 nova nova  97 Feb 20  2017 .
drwxr-xr-x. 9 nova nova 111 Feb 17  2017 ..
drwxr-xr-x. 2 nova nova  54 Feb 17  2017 _base
-rw-r--r--. 1 nova nova  44 May 26  2017 compute_nodes
drwxr-xr-x. 2 nova nova  54 Feb 17  2017 e560e649-41fd-46a2-a3d2-5f4750ba2bb4
drwxr-xr-x. 2 nova nova  93 Feb 17  2017 locks

/var/lib/nova/instances/_base:
total 18176
drwxr-xr-x. 2 nova nova       54 Feb 17  2017 .
drwxr-xr-x. 5 nova nova       97 Feb 20  2017 ..
-rw-r--r--. 1 qemu qemu 41126400 May 26  2017 faf1184c098da91e90290a920b8fab1ee6e1d4c4

/var/lib/nova/instances/e560e649-41fd-46a2-a3d2-5f4750ba2bb4:
total 2104
drwxr-xr-x. 2 nova nova      54 Feb 17  2017 .
drwxr-xr-x. 5 nova nova      97 Feb 20  2017 ..
-rw-r--r--. 1 qemu qemu   48957 Feb 20  2017 console.log
-rw-r--r--. 1 qemu qemu 2097152 Feb 20  2017 disk
-rw-r--r--. 1 nova nova      79 Feb 17  2017 disk.info

/var/lib/nova/instances/locks:
total 0
drwxr-xr-x. 2 nova nova 93 Feb 17  2017 .
drwxr-xr-x. 5 nova nova 97 Feb 20  2017 ..
-rw-r--r--. 1 nova nova  0 Feb 17  2017 nova-faf1184c098da91e90290a920b8fab1ee6e1d4c4
-rw-r--r--. 1 nova nova  0 Feb 17  2017 nova-storage-registry-lock
'''.strip()


def test_ls_var_lib_nova_instances():
    ls_var_lib_nova_instances = ls_instances.LsVarLibNovaInstances(context_wrap(LS_VAR_LIB_NOVA_INSTANCES))
    assert ls_var_lib_nova_instances.dirs_of('/var/lib/nova/instances/') == ['.', '..', '11415c6c-a2a5-45f0-a198-724246b96631', '_base', 'locks']
    assert ls_var_lib_nova_instances.listings['/var/lib/nova/instances/11415c6c-a2a5-45f0-a198-724246b96631']['entries']['console.log']['se_type'] == 'nova_var_lib_t'
    assert ls_var_lib_nova_instances.dir_entry('/var/lib/nova/instances/locks', 'nova-storage-registry-lock') == {'se_type': 'nova_var_lib_t', 'name': 'nova-storage-registry-lock', 'perms': 'rw-r--r--.', 'se_user': 'system_u', 'raw_entry': '-rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 nova-storage-registry-lock', 'se_mls': 's0', 'se_role': 'object_r', 'owner': 'nova', 'group': 'nova', 'type': '-', 'dir': '/var/lib/nova/instances/locks'}
    assert ls_var_lib_nova_instances.dir_entry('/var/lib/nova/instances/11415c6c-a2a5-45f0-a198-724246b96631', 'console.log')['owner'] == 'root'


def test_ls_r_var_lib_nova_instances():
    ls_r_instances = ls_instances.LsRVarLibNovaInstances(context_wrap(LS_R_VAR_LIB_NOVA_INSTANCES))
    assert ls_r_instances.dir_entry('/var/lib/nova/instances/e560e649-41fd-46a2-a3d2-5f4750ba2bb4', 'console.log')['size'] == 48957


def test_ls_var_lib_nova_instances_doc_examples():
    failed, total = doctest.testmod(
        ls_instances,
        globs={'ls_var_lib_nova_instances': ls_instances.LsVarLibNovaInstances(context_wrap(LS_VAR_LIB_NOVA_INSTANCES)),
               'ls_r_var_lib_nova_instances': ls_instances.LsRVarLibNovaInstances(context_wrap(LS_R_VAR_LIB_NOVA_INSTANCES))}
    )
    assert failed == 0



"""
KSMState - file ``/sys/kernel/mm/ksm/run``
==========================================

Parser to get the kernel samepage merging state by reading the file
``/sys/kernel/mm/ksm/run``.
"""
from insights.core import Parser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.ksmstate)
class KSMState(Parser):
    """
    Parser to get the kernel samepage merging state by reading the file
    ``/sys/kernel/mm/ksm/run``.

    Typical output of ``/sys/kernel/mm/ksm/run`` likes:

        0

    From https://www.kernel.org/doc/Documentation/vm/ksm.txt::

        set 0 to stop ksmd from running but keep merged pages,
        set 1 to run ksmd e.g. "echo 1 > /sys/kernel/mm/ksm/run",
        set 2 to stop ksmd and unmerge all pages currently merged, but leave
              mergeable areas registered for next run

    Examples:
        >>> ksm.value == '0'
        True
        >>> ksm.is_running
        False

    Raises:
        SkipComponent: when input is empty.
        ParseException: when the content is not expected.
    """
    def parse_content(self, content):
        if not content:
            raise SkipComponent
        if len(content) != 1 or content[0] not in ('0', '1', '2'):
            raise ParseException("Incorrect content: '{0}'".format(content))

        self.value = content[0]
        self.is_running = self.value == '1'



import doctest

from insights.core.filters import add_filter
from insights.parsers import ls_usr_sbin
from insights.parsers.ls_usr_sbin import LsUsrSbin
from insights.specs import Specs
from insights.tests import context_wrap

LS_USR_SBIN = """
total 41472
-rwxr-xr-x. 1 0  0   11720 Mar 18  2014 accessdb
-rwxr-xr-x. 1 0  0    3126 Oct  4  2013 addgnupghome
-rwxr-xr-x. 1 0  0   20112 Jun  1  2017 addpart
-rwxr-xr-x. 1 0  0  371912 Jan 27  2014 postconf
-rwxr-sr-x. 1 0 90  218552 Jan 27  2014 postdrop
"""


def test_ls_usr_sbin():
    ls_usr_sbin = LsUsrSbin(context_wrap(LS_USR_SBIN, path='insights_commands/ls_-ln_.usr.sbin'))
    assert ls_usr_sbin.files_of('/usr/sbin') == ['accessdb', 'addgnupghome', 'addpart', 'postconf', 'postdrop']
    postdrop = ls_usr_sbin.dir_entry('/usr/sbin', 'postdrop')
    assert postdrop is not None
    assert postdrop == {
        'group': '90',
        'name': 'postdrop',
        'links': 1,
        'perms': 'rwxr-sr-x.',
        'raw_entry': '-rwxr-sr-x. 1 0 90  218552 Jan 27  2014 postdrop',
        'owner': '0',
        'date': 'Jan 27  2014',
        'type': '-',
        'size': 218552,
        'dir': '/usr/sbin'}


def test_ls_usr_sbin_doc_examples():
    env = {
        'Specs': Specs,
        'add_filter': add_filter,
        'LsUsrSbin': LsUsrSbin,
        'ls_usr_sbin': LsUsrSbin(context_wrap(LS_USR_SBIN, path='insights_commands/ls_-ln_.usr.sbin')),
    }
    failed, total = doctest.testmod(ls_usr_sbin, globs=env)
    assert failed == 0



from insights.util import deprecated


class BlacklistedSpec(Exception):
    """
    Exception to be thrown when a blacklisted spec is found.
    """
    pass


class CalledProcessError(Exception):
    """
    Raised if call fails.

    Parameters:
        returncode (int): The return code of the process executing the command.
        cmd (str): The command that was executed.
        output (str): Any output the command produced.
    """

    def __init__(self, returncode, cmd, output=None):
        self.returncode = returncode
        self.cmd = cmd
        self.output = output
        super(CalledProcessError, self).__init__(returncode, cmd, output)

    def __unicode__(self):
        name = self.__class__.__name__
        rc = self.returncode
        cmd = self.cmd
        output = self.output
        return '<{}({}, {!r}, {!r})>'.format(name, rc, cmd, output)


class InvalidArchive(Exception):
    def __init__(self, msg):
        super(InvalidArchive, self).__init__(msg)
        self.msg = msg


class MissingRequirements(Exception):
    """
    Raised during evaluation if a component's dependencies aren't met.
    """
    def __init__(self, requirements):
        self.requirements = requirements
        super(MissingRequirements, self).__init__(requirements)


class ParseException(Exception):
    """
    Exception that should be thrown from parsers that encounter
    exceptions they recognize while parsing. When this exception
    is thrown, the exception message and data are logged and no
    parser output data is saved.
    """
    pass


class SkipComponent(Exception):
    """
    This class should be raised by components that want to be taken out of
    dependency resolution.
    """
    pass


class TimeoutException(Exception):
    """ Raised whenever a :class:`datasource` hits the set timeout value. """
    pass


class ValidationException(Exception):
    def __init__(self, msg, r=None):
        if r:
            msg = "%s: %s" % (msg, r)
        super(ValidationException, self).__init__(msg)


class ContentException(SkipComponent):
    """ Raised whenever a :class:`datasource` fails to get data. """
    pass


class InvalidContentType(InvalidArchive):
    def __init__(self, content_type):
        self.msg = 'Invalid content type: "%s"' % content_type
        super(InvalidContentType, self).__init__(self.msg)
        self.content_type = content_type


class SkipException(SkipComponent):
    """
    .. warning::
        This class is deprecated, please use
        :py:class:`insights.core.exceptions.SkipComponent` instead.

    Exception that should be thrown from parsers that are explicitly
    written to look for errors in input data.  If the expected error
    is not found then the parser should throw this exception to
    signal to the infrastructure that the parser's output should not be
    retained.
    """
    def __init__(self, msg):
        deprecated(
            SkipException,
            "Please use the :class:`insights.core.exceptions.SkipComponent` instead.",
            "3.3.0"
        )
        super(SkipException, self).__init__(msg)



"""
RpmVPackages - command ``/bin/rpm -V <packages>``
=================================================

Below packages are verified:
    - coreutils
    - procps
    - procps-ng
    - shadow-utils
    - passwd
    - sudo
    - chrony
"""

from insights.core import CommandParser
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.rpm_V_packages)
class RpmVPackages(CommandParser):
    """
    Class for parsing ``/bin/rpm -V <packages>`` command.

    Attributes:
        packages_list (list of dictionaries): every dictionary contains information about one entry

    Sample output of this command is::

        package procps is not installed
        ..?......  c /etc/sudoers
        ..?......    /usr/bin/sudo
        ..?......    /usr/bin/sudoreplay
        missing     /var/db/sudo/lectured (Permission denied)

    Examples:
        >>> type(rpm_v_packages)
        <class 'insights.parsers.rpm_v_packages.RpmVPackages'>
        >>> len(rpm_v_packages.packages_list)
        5
        >>> sorted(rpm_v_packages.packages_list[0].items())
        [('attributes', None), ('file', None), ('line', 'package procps is not installed'), ('mark', None)]
        >>> sorted(rpm_v_packages.packages_list[1].items())
        [('attributes', '..?......'), ('file', '/etc/sudoers'), ('line', '..?......  c /etc/sudoers'), ('mark', 'c')]
"""

    def parse_content(self, content):
        self.packages_list = []

        for line in content:
            line_parts = line.split()
            if "package" in line_parts[0] or "missing" in line_parts[0]:
                entry = {"line": line.strip(), "attributes": None, "mark": None, "file": None}
            elif len(line_parts) == 3:
                entry = {"line": line.strip(), "attributes": line_parts[0], "mark": line_parts[1],
                         "file": line_parts[2]}
            else:
                entry = {"line": line.strip(), "attributes": line_parts[0], "mark": None,
                         "file": line_parts[1]}
            self.packages_list.append(entry)



import pytest

from insights.core.exceptions import ParseException
from insights.parsers.scsi import SCSI
from insights.tests import context_wrap


SCSI_OUTPUT = """
Attached devices:
Host: scsi0 Channel: 03 Id: 00 Lun: 00
  Vendor: HP       Model: P420i            Rev: 3.54
  Type:   RAID                             ANSI  SCSI revision: 05
Host: scsi0 Channel: 00 Id: 00 Lun: 00
  Vendor: HP       Model: LOGICAL VOLUME   Rev: 3.54
  Type:   Direct-Access                    ANSI  SCSI revision: 05
Host: scsi0 Channel: 00 Id: 00 Lun: 01
  Vendor: HP       Model: LOGICAL VOLUME   Rev: 3.54
  Type:   Direct-Access                    ANSI  SCSI revision: 05
Host: scsi0 Channel: 00 Id: 00 Lun: 02
  Vendor: HP       Model: LOGICAL VOLUME   Rev: 3.54
  Type:   Direct-Access                    ANSI  SCSI revision: 05
Host: scsi0 Channel: 00 Id: 00 Lun: 03
  Vendor: HP       Model: LOGICAL VOLUME   Rev: 3.54
  Type:   Direct-Access                    ANSI  SCSI revision: 05
"""

SCSI_OUTPUT_SINGLE_SPACED_ANSI_SCSI = """
Attached devices:
Host: scsi0 Channel: 00 Id: 00 Lun: 00
  Vendor: VMware   Model: Virtual disk     Rev: 1.0
  Type:   Direct-Access                    ANSI SCSI revision: 02
Host: scsi0 Channel: 00 Id: 01 Lun: 00
  Vendor: VMware   Model: Virtual disk     Rev: 1.0
  Type:   Direct-Access                    ANSI SCSI revision: 02
Host: scsi0 Channel: 00 Id: 02 Lun: 00
  Vendor: VMware   Model: Virtual disk     Rev: 1.0
  Type:   Direct-Access                    ANSI SCSI revision: 02
Host: scsi0 Channel: 00 Id: 03 Lun: 00
  Vendor: VMware   Model: Virtual disk     Rev: 1.0
  Type:   Direct-Access                    ANSI SCSI revision: 02
"""

SCSI_OUTPUT_MISSING_HEADER = """
Host: scsi0 Channel: 03 Id: 00 Lun: 00
  Vendor: HP       Model: P420i            Rev: 3.54
  Type:   RAID                             ANSI  SCSI revision: 05
"""

SCSI_OUTPUT_INNORMAL_VEMDOR_LINE = """
Attached devices:
Host: scsi0 Channel: 03 Id: 00 Lun: 00
  Vendor: HP       Model: P420i            Rev: 3.:
  Type:   RAID                             ANSI  SCSI revision: 05
"""

SCSI_OUTPUT_EMPTY = """
"""


def test_parse():
    context = context_wrap(SCSI_OUTPUT)
    result = SCSI(context)
    assert len(result) == 5
    r = result[0]
    assert r.host == "scsi0"
    assert r.get('host') == 'scsi0'
    assert r.channel == "03"
    assert r.id == "00"
    assert r.lun == "00"
    assert r.vendor == "HP"
    assert r.model == "P420i"
    assert r.rev == "3.54"
    assert r.type == "RAID"
    assert r.ansi__scsi_revision == "05"

    r = result[1]
    assert r.model == "LOGICAL VOLUME"

    r = result[4]
    assert r.type == "Direct-Access"

    assert [disc.vendor for disc in result] == ['HP'] * 5


def test_single_spaced_ansi_scsi():
    result = SCSI(context_wrap(SCSI_OUTPUT_SINGLE_SPACED_ANSI_SCSI))
    assert len(result) == 4


def test_missing_header():
    with pytest.raises(ParseException) as excinfo:
        result = SCSI(context_wrap(SCSI_OUTPUT_MISSING_HEADER))
        assert result is None
    assert 'Expected Header: Attached devices:' in str(excinfo.value)


def test_missing_innormal_rev():
    with pytest.raises(ParseException) as excinfo:
        result = SCSI(context_wrap(SCSI_OUTPUT_INNORMAL_VEMDOR_LINE))
        assert result is None
    assert 'Parse error for current line:' in str(excinfo.value)


def test_empty():
    with pytest.raises(ParseException) as excinfo:
        result = SCSI(context_wrap(SCSI_OUTPUT_EMPTY))
        assert result is None
    assert 'Empty content of file /proc/scsi/scsi' in str(excinfo.value)



"""
BlockIDInfo - command ``blkid``
===============================

This module provides the class ``BlockIDInfo`` which processes ``blkid`` command
output.  Typical output looks like::

    /dev/sda1: UUID="3676157d-f2f5-465c-a4c3-3c2a52c8d3f4" TYPE="xfs"
    /dev/sda2: UUID="UVTk76-UWOc-vk7s-galL-dxIP-4UXO-0jG4MH" TYPE="LVM2_member"
    /dev/mapper/rhel_hp--dl160g8--3-root: UUID="11124c1d-990b-4277-9f74-c5a34eb2cd04" TYPE="xfs"
    /dev/mapper/rhel_hp--dl160g8--3-swap: UUID="c7c45f2d-1d1b-4cf0-9d51-e2b0046682f8" TYPE="swap"
    /dev/mapper/rhel_hp--dl160g8--3-home: UUID="c7116820-f2de-4aee-8ea6-0b23c6491598" TYPE="xfs"
    /dev/mapper/rhel_hp--dl160g8--3-lv_test: UUID="d403bcbd-0eea-4bff-95b9-2237740f5c8b" TYPE="ext4"
    /dev/cciss/c0d1p3: LABEL="/u02" UUID="004d0ca3-373f-4d44-a085-c19c47da8b5e" TYPE="ext3"
    /dev/cciss/c0d1p2: LABEL="/u01" UUID="ffb8b27e-5a3d-434c-b1bd-16cb17b0e325" TYPE="ext3"
    /dev/loop0: LABEL="Satellite-5.6.0 x86_64 Disc 0" TYPE="iso9660"
    /dev/block/253:1: UUID="f8508c37-eeb1-4598-b084-5364d489031f" TYPE="ext3"

The class has one attribute ``data`` which is a ``list`` representing each line
of the input data as a ``dict`` with keys corresponding to the keys in the
output.

Examples:
    >>> block_id = shared[BlockIDInfo]
    >>> block_id.data[0]
    {'NAME': '/dev/sda1', 'UUID': '3676157d-f2f5-465c-a4c3-3c2a52c8d3f4', 'TYPE': 'xfs'}
    >>> block_id.data[0]['TYPE']
    'xfs'
    >>> block_id.filter_by_type('ext3')
    [{'NAME': '/dev/cciss/c0d1p3', 'LABEL': '/u02', 'UUID': '004d0ca3-373f-4d44-a085-c19c47da8b5e',
      'TYPE': 'ext3'},
     {'NAME': '/dev/block/253:1', 'UUID': 'f8508c37-eeb1-4598-b084-5364d489031f','TYPE': 'ext3'},
     {'NAME': '/dev/cciss/c0d1p2', 'LABEL': '/u01', 'UUID': 'ffb8b27e-5a3d-434c-b1bd-16cb17b0e325',
      'TYPE': 'ext3'}]

"""
import re
from .. import parser, CommandParser
from insights.specs import Specs


@parser(Specs.blkid)
class BlockIDInfo(CommandParser):
    """Class to process the ``blkid`` command output.

    Attributes:
        data (list): A list containing a dictionary for each line of the output in
            the form::

                [
                    {
                        'NAME': "/dev/sda1"
                        'UUID': '3676157d-f2f5-465c-a4c3-3c2a52c8d3f4',
                        'TYPE': 'xfs'
                    },
                    {
                        'NAME': "/dev/cciss/c0d1p3",
                        'LABEL': '/u02',
                        'UUID': '004d0ca3-373f-4d44-a085-c19c47da8b5e',
                        'TYPE': 'ext3'
                    }
                ]
    """
    def parse_content(self, content):
        blkid_output = []
        for line in (l for l in content if l.strip()):
            dev_name, attributes = line.rsplit(":", 1)
            device = dict((k, v) for k, v in re.findall(r'(\S+)=\"(.*?)\"\s?', line))
            device['NAME'] = dev_name.strip()
            blkid_output.append(device)

        self.data = blkid_output
        """list of dict: List containing dict for each line of command output."""

    def filter_by_type(self, fs_type):
        """list: Returns a list of all entries where TYPE = ``fs_type``."""
        return [row for row in self.data if row.get('TYPE') == fs_type]
